/*********************************************************************
* This software module was originally developed by 
*   Trac D. Tran, Lijie Liu, Pankaj N. Topiwala, and 
*   Patrick Rault (FastVdo LLC: "FastVDO"), 
* and modified by
*   Shawn (Sheng) Zhong (Broadcom Corporation: "Broadcom"),
*   Yuriy A. Reznik (QUALCOMM Incorporated: "QUALCOMM"),
*   Cixun Zhang (Zhejiang University: "ZJU"), and
*   Arianne T. Hinds (International Business Machines Corporation: "IBM")
* in the course of development of the ISO/IEC 23002-1 "accuracy requirements 
* for implementations of integer-output 8x8 IDCT" standard for reference 
* purposes and its performance may not have been optimized. This software 
* module is an implementation of one or two more tools as specified by the 
* ISO/IEC 23002-1 standard.
*
* ISO/IEC gives You a royalty-free, worldwide, non-exclusive, copyright 
* license to copy, distribute, and make derivative works of this software 
* module or modifications thereof for use in implementations of the 
* ISO/IEC 23002-1 standard in products that satisfy conformance criteria 
* (if any).
*
* Those intending to use this software module in products are advised that 
* its use may infringe existing patents. ISO/IEC have no liability for use 
* of this software module or modifications thereof.
*
* Copyright is not released for products that do not conform to audiovisual 
* and image-coding related ITU Recommendations and/or ISO/IEC International 
* Standards.
*
******** Begin section to be removed when relevant standard is published *****
*
* Assurance that the originally developed software module can be used
* (1) in the ISO/IEC 23002-1 standard once the ISO/IEC 23002-1
*     standard has been adopted; and
* (2) to develop the ISO/IEC 23002-1 standard:
*
* FastVDO, Broadcom, QUALCOMM, ZJU, and IBM each individually grant ISO/IEC 
* all rights necessary to include the originally developed software module or 
* modifications thereof in the ISO/IEC 23002-1 standard and to permit 
* ISO/IEC to offer You a royalty-free, worldwide, non-exclusive, copyright 
* license to copy, distribute, and make derivative works for use in 
* implementations of the ISO/IEC 23002-1 standard in products that satisfy 
* conformance criteria (if any), and to the extent that such originally 
* developed software module or portions of it are included in the ISO/IEC 
* 23002-1 standard. To the extent that FastVDO, Broadcom, QUALCOMM, ZJU, and IBM, 
* own patent rights that would be required to make, use, or sell the originally 
* developed software module or portions thereof included in the ISO/IEC 
* 23002-1 standard in a conforming product, FastVDO, Broadcom, QUALCOMM, ZJU, 
* and IBM each individually will assure the ISO/IEC that they are willing to 
* negotiate licenses under reasonable and non-discriminatory terms and 
* conditions with applicants throughout the world.
*
* ISO/IEC gives You a free license to this software module or modifications
* thereof for the sole  purpose of developing the ISO/IEC 23002-1 standard.
*
******** end of section to be removed when the standard is published *****
*
* FastVDO, Broadcom, QUALCOMM, ZJU, and IBM retain full rights to modify and 
* use the code for their own purposes, assign or donate the code to a third 
* party and to inhibit third parties from using the code for products that do 
* not conform to MPEG-related ITU Recommendations and/or ISO/IEC International 
* Standards.
*
* This copyright notice must be included in all copies or derivative works.
* Copyright (c) ISO/IEC 2006.
**********************************************************************/

#include "testbed.h"

/* pixel bit-depth: */
int B = 8;

int clip3 (int x, int y, int z)
{
  return (z < x)? x: ((z > y)? y: z);
}

/* Pseudo-random generator specified by ISO/IEC 23002-1 */
long randx = 1;
long rand_gen (long L, long H)
{
  long i, j;
  double x;

  randx = (randx * 1103515245) + 12345;
  i = randx & 0x7ffffffe;
  x = ((double) i) / ((double) 0x7fffffff);
  x *= (L+H+1);
  j = (long) x;
  return j - L;
}

void rand_gen_reset (void)
{
  randx = 1;
}

/* The cosine lookup table */
/* coslu[a][b] = C(b)/2 * cos[(2a+1)b*pi/16] */
double coslu[8][8];

/* Routine to initialise the cosine lookup table */
void dct_init (void)
{
  int a, b;
  double tmp;

  for (a=0; a<8; a++) {
    for (b=0; b<8; b++) {
      tmp = cos((double)((a+a+1)*b) * (3.14159265358979323846 / 16.0));
      if (b==0)
        tmp /= sqrt(2.0);
      coslu[a][b] = tmp * 0.5;
    }
  }
}

/* Reference integer-valued forward transform: */
void ref_fdct (short *block8x8)
{
  int x, y, u, v;
  double tmp, tmp2, e;
  double res[8][8];

  for (v=0; v<8; v++) 
  {
    for (u=0; u<8; u++) 
    {
      tmp = 0.0;
      for (y=0; y<8; y++) 
      {
        tmp2 = 0.0;
        for (x=0; x<8; x++) 
        {
          tmp2 += (double) block8x8[y*8+x] * coslu[x][u];
        }
        tmp += coslu[y][v] * tmp2;
      }
      res[v][u] = tmp;
    }
  }

  /* round results to integers: */
  for (v=0; v<8; v++) 
  {
    for (u=0; u<8; u++) 
    {		
      e = 0.0; 
      if (u==0 && (v==0 || v==4)) e = 10E-10; // eps value as defined in Sec 5.2 
      if (u==4 && (v==0 || v==4)) e = 10E-10;
      tmp = res[v][u];
      if (tmp < 0.0) 
        x = - ((int) (e + 0.5 - tmp));
      else 
        x = (int) (tmp + 0.5 + e);
      block8x8[v*8+u] = (int) x;
    }
  }
}

/* Reference floating-point forward transform: */
void ref_double_fdct (double block[64])
{
  int x,y,u,v;
  double tmp, tmp2;
  double res[8][8];

  for (v=0; v<8; v++) 
  {
    for (u=0; u<8; u++) 
    {
      tmp = 0.0;
      for (y=0; y<8; y++) 
      {
        tmp2 = 0.0;
        for (x=0; x<8; x++) 
        {
          tmp2 += (double) block[y*8+x] * coslu[x][u];
        }
        tmp += coslu[y][v] * tmp2;
      }
      res[v][u] = tmp;
    }
  }

  for (v=0; v<8; v++) 
  {
    for (u=0; u<8; u++) 		
    {
      block[v*8+u] = res[v][u];
    }
  }
}

/* Reference integer-valued inverse transform  */
void ref_idct (short * block8x8)
{
  int x, y, u, v;
  double tmp, tmp2;
  double res[8][8];

  for (y=0; y<8; y++) 
  {
    for (x=0; x<8; x++) 
    {
      tmp = 0.0;
      for (v=0; v<8; v++) 
      {
        tmp2 = 0.0;
        for (u=0; u<8; u++) 
        {
          tmp2 += (double) block8x8[v*8+u] * coslu[x][u];
        }
        tmp += coslu[y][v] * tmp2;
      }
      res[y][x] = tmp;
    }
  }

  for (v=0; v<8; v++) 
  {
    for (u=0; u<8; u++) 
    {
      tmp = res[v][u];
      if (tmp < 0.0) 
        x = - ((int) (0.5 - tmp));
      else 
        x = (int) (tmp + 0.5);
      block8x8[v*8+u] = (short) x;
    }
  }
}

/* Reference double-precision inverse transform  */
void ref_double_idct (double *block)
{
  int x, y, u, v;
  double tmp, tmp2;
  double res[8][8];

  for (y=0; y<8; y++) 
  {
    for (x=0; x<8; x++) 
    {
      tmp = 0.0;
      for (v=0; v<8; v++) 
      {
        tmp2 = 0.0;
        for (u=0; u<8; u++) 
        {
          tmp2 += block[v*8+u] * coslu[x][u];
        }
        tmp += coslu[y][v] * tmp2;
      }
      res[y][x] = tmp;
    }
  }

  for (v=0; v<8; v++) 
  {
    for (u=0; u<8; u++) 		
    {
      block[v*8+u] = res[v][u];
    }
  }
}

char *meets (double val, double limit)
{
  return ((fabs(val) <= limit) ? "meets" : "FAILS");
}

/* [5.5.1.1] IDCT test procedure using pseudo-random input: */
void idct_test (long minpix, long maxpix, long sign, long niters, FILE*  outputfile)
{		
  double max[NPROPOSAL], total[NPROPOSAL], pk[NPROPOSAL], err;
  short  block[64];	    /* random source data */
  short  refcoefs[64];    /* coefs from reference FDCT */
  short  refout[64];  	/* output from reference IDCT */
  short  testout[64];     /* output from test IDCT */
  long   sumerrs[NPROPOSAL][64], sumsqerrs[NPROPOSAL][64];
  long   maxerr[NPROPOSAL][64];
  int    proposal, nproposal = NPROPOSAL ;
  int    i, curiter, dct_range, idct_range;

  idct_range = (1<<B);     // 2^B
  dct_range =  (1<<(B+3)); // 2^(B+3)

  for (proposal=0; proposal<nproposal; proposal++)
  {
    for (i = 0; i < 64; i++)
      sumerrs[proposal][i] = sumsqerrs[proposal][i] = maxerr[proposal][i] = 0;
  }

  /* reset random number generator: */
  rand_gen_reset ();

  /* Loop once per generated random-data block */
  for (curiter = 0; curiter < niters; curiter++) 
  {
    /* generate a pseudo-random block of data */
    for (i = 0; i < 64; i++)
      block[i] = (short) (rand_gen(-minpix,maxpix) * sign);

    /* perform reference FDCT */
    memcpy(refcoefs, block, sizeof(short)*64);
    ref_fdct(refcoefs);

    /* clip */
    for (i = 0; i < 64; i++) 
      refcoefs[i] = clip3(-dct_range, dct_range-1, refcoefs[i]);

    /* perform reference IDCT */
    memcpy(refout, refcoefs, sizeof(short)*64);
    ref_idct(refout);

    /* clip */
    for (i = 0; i < 64; i++) 
      refout[i] = clip3(-idct_range, idct_range-1, refout[i]);

    /* perform test IDCT */
    for (proposal=0; proposal < nproposal; proposal++)
    { 
      memcpy(testout, refcoefs, sizeof(short)*64);  
      idct_proposal[proposal](testout) ; 
      /* clip */
      for (i = 0; i < 64; i++) 
        testout[i] = clip3(-idct_range, idct_range-1, testout[i]) ;
      /* accumulate error stats */
      for (i = 0; i < 64; i++) 
      {
        register int err = testout[i] - refout[i];
        sumerrs[proposal][i] += err;
        sumsqerrs[proposal][i] += err * err;
        if (err < 0) err = -err;
        if (maxerr[proposal][i] < err) maxerr[proposal][i] = err;
      }
    }
    if (curiter % 100 == 99) 
    {
      fprintf(stderr, ".");
      fflush(stderr);
    }
  }

  /* print header: */
  fprintf(outputfile, "\n");
  fprintf(outputfile, "IDCT test\n");
  fprintf(outputfile, "Test conditions: -L = %ld, +H = %ld, sign = %ld, #iters = %ld \n",
    minpix, maxpix, sign, niters);
  fprintf(outputfile, "\nProposals: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    max[proposal] = total[proposal] = pk[proposal] = 0 ;
    fprintf(outputfile, "%s, ", proposal_label[proposal]);
  }

  /* peak errors: */
  fprintf(outputfile, "\n\nPeak absolute values of errors:\n");
  for (i=0; i<64; i++)
  {
    fprintf(outputfile, "%2d: ", i) ;
    for (proposal = 0; proposal < nproposal; proposal++)
    {
      fprintf(outputfile, "%2d, ", maxerr[proposal][i]); // Reznik, Yuriy 12/07/2005
      if (pk[proposal] < maxerr[proposal][i]) 
        pk[proposal] = maxerr[proposal][i];
    }
    fprintf(outputfile, "\n");	
  }
  fprintf(outputfile, "\nWorst peak errors: ") ;	
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%.2e, ", pk[proposal]);		
  }
  fprintf(outputfile, "\n<=1.0: ");	
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", meets((double) pk[proposal], 1.0));		
  }

  /* mse errors: */
  fprintf(outputfile, "\n\nMean square errors:\n");	
  for (i = 0; i < 64; i++) 
  {
    fprintf(outputfile, "%2d: ", i) ;	
    for (proposal = 0; proposal < nproposal; proposal++)
    {	
      double err = (double) sumsqerrs[proposal][i]  / ((double) niters);
      if (max[proposal]<err)
        max[proposal] = err ;
      total[proposal] += (double) sumsqerrs[proposal][i];
      fprintf(outputfile, "%.2e, ", ABS(err));		
    }
    fprintf(outputfile, "\n");
  }
  fprintf(outputfile, "\nWorst pmse: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%.2e, ", max[proposal]);
  }
  /* 09/12/06: YR: added check for max pmse limit: */
  fprintf(outputfile, "\n<=0.06: ") ;	
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", meets((double) max[proposal], 0.06));		
    max[proposal]=0 ;
  }
  /*******/
  fprintf(outputfile, "\n\nOverall pmse: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    total[proposal] /= (double) (64*niters);
    fprintf(outputfile, "%.2e,", total[proposal]);
  }
  fprintf(outputfile,"\n<=0.02: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", meets(total[proposal], 0.02));
    total[proposal] = 0 ;
  }

  /* mean errors: */
  fprintf(outputfile, "\n\nOverall Mean errors:\n");
  for (i = 0; i < 64; i++) 
  {
    fprintf(outputfile, "%2d: ", i) ;
    for (proposal = 0; proposal < nproposal; proposal++)
    {
      err = (double) sumerrs[proposal][i] / ((double) niters);
      err = ABS(err) ;
      total[proposal] += (double) sumerrs[proposal][i];
      fprintf(outputfile, "%.2e, ", err);
      if (max[proposal] < err) max[proposal] = err;
    }
    fprintf(outputfile, "\n");
  }
  fprintf(outputfile, "\nWorst Mean errors: ") ;
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%.2e, ", max[proposal]);
  }
  /* 09/12/06: YR: added check for max pme limit: */
  fprintf(outputfile, "\n<=0.015: ") ;	
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", meets((double) max[proposal], 0.015));		
    max[proposal] = 0 ;
  }
  /*******/
  fprintf(outputfile, "\n\nOverall Mean errors: ") ;
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    total[proposal] /= (double) (64*niters);
    fprintf(outputfile, "%.2e, ", total[proposal]);
  }
  fprintf(outputfile, "\n<=0.0015: ") ;
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ",meets(total[proposal], 0.0015));
    total[proposal] = 0 ;
  }
  fprintf(outputfile, "\n") ;
  return;
}

/* [5.5.2] Test of all-zero input behavior: */
void zeroinput_test (FILE* outputfile)
{
  short testout[64], i, j; 
  int proposal, nproposal = NPROPOSAL;

  fprintf(outputfile, "\n\nZero input test: ") ;
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    memset(testout, 0, sizeof(short)*64);
    idct_proposal[proposal](testout);
    for (i = 0, j=0; i < 64; i++) 
    {
      if (testout[i])
      {
        fprintf(outputfile, "FAILS, ");
        j++ ;
        break;
      }
    }
    if (!j) 
    {
      fprintf(outputfile, "meets, ");	
    }
  }
  fprintf(outputfile, "\n");
}

/* [A.2] Test of near-DC inversion behaviour: */
void neardc_test(FILE* outputfile)
{
  short ref[64];
  short refout[64];  	// output from reference IDCT 
  short testout[64];  // output from test IDCT 
  int pr[NPROPOSAL][64];
  int proposal, nproposal = NPROPOSAL;
  int dct_range = 1 << (B+3); 
  int idct_range = 1 << B; 
  int k, l; 

  memset(pr, 0, sizeof(int)*nproposal*64);
  for (k=0; k < dct_range*2; k++)		
  {
    memset(ref, 0, sizeof(short)*64); 
    ref[0] = k - dct_range;
    if ((k & 1) == 0) 
      ref[63] = 1;
    memcpy(refout, ref, sizeof(short)*64); 
    ref_idct(refout);
    for(l=0; l<64; l++)		
      refout[l] = clip3(-idct_range, idct_range-1, refout[l]);
    /* test proposals: */
    for (proposal=0; proposal < nproposal; proposal++)
    {
      memcpy(testout, ref, sizeof(short)*64); 
      idct_proposal[proposal](testout);
      for (l=0; l<64; l++)	
      {
        testout[l] = clip3(-idct_range, idct_range-1, testout[l]);
        if (pr[proposal][l] < abs(testout[l] - refout[l]))
          pr[proposal][l] = abs(testout[l] - refout[l]);
      }
    }
  }

  fprintf(outputfile, "\n\nNear DC test:\n") ;
  for(l=0; l<64; l++)		
  {
    fprintf(outputfile,"%2d: ",l) ;	
    for (proposal = 0; proposal < nproposal; proposal++)
    {
      fprintf(outputfile, "%d, ",pr[proposal][l]);		
    }
    fprintf(outputfile, "\n") ;		
  }
  /* additional pass/fail printouts: */
  fprintf(outputfile, "\n") ;		
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    for(l=0; l<64; l++)	
    {
      if (pr[proposal][l] > 1) 
      {
        fprintf(outputfile, "fail, ");
        goto skip;
      }
    }
    fprintf(outputfile, "pass, ");
    skip:;
  }

}

/* IDCT linearity test 
* -- added by Cixun Zhang, ZJU */
void linearity_test (FILE* outputfile)
{
  short input[64];
  short input_inv[64];

  int lin_pass[NPROPOSAL];	
  int proposal, nproposal = NPROPOSAL ;

  int idct_range = (1<<8); 
  int level, run;
  int i, k; 

  for (proposal = 0; proposal<nproposal; proposal++)
    lin_pass[proposal] = 1;

  for (proposal = 0; proposal<nproposal; proposal++)
  {
    for (level=1; level<528; level+=2)
    {
      for (run=0; run<64; run++)
      {
        memset(input, 0, sizeof(input));
        memset(input_inv, 0, sizeof(input_inv));

        input[run] = level;
        input_inv[run] = -level;

        idct_proposal[proposal](input);
        for (k=0; k<64; k++)
          input[k] = clip3(-idct_range, idct_range-1, input[k]);
        idct_proposal[proposal](input_inv);
        for (k=0; k<64; k++)
          input_inv[k] = clip3(-idct_range, idct_range-1, input_inv[k]);

        for (i=0; i<64; i++)
        {
          if(input[i] + input_inv[i] != 0)
          {
            lin_pass[proposal] = 0;
            goto NEXT;
          }
        }
      }
    }
NEXT: ;
  }
  fprintf(outputfile, "\n\nLinearity test: ") ;
  for (proposal = 0  ; proposal<nproposal ; proposal++)
  {
    if (lin_pass[proposal])
      fprintf(outputfile, "pass, ");
    else
      fprintf(outputfile, "fail, ");
  }
  fprintf(outputfile, "\n") ;		

  return;
}


/* number of raised/dropped bits in invertibility test: */
int scale = 0;

// Invertibility test
// -- added by Arianne T. Hinds, IBM
void invert_test (long L, long H, long sign, long run, FILE* outputfile)
{
  int i, j, k;
  int mismatch = 0;
  short spatial_data [64];
  short alg_data [NPROPOSAL][64];
  short out_data [NPROPOSAL][64];
  int delta [NPROPOSAL][64];
  int mismatches [NPROPOSAL];

  memset(mismatches, 0, NPROPOSAL * sizeof(int)); 
  for (i=0; i<NPROPOSAL; i++) 
    memset(delta[i], 0, 64 * sizeof(int));

  /* reset random number generator: */
  rand_gen_reset ();

  for (j=0; j<run; j++) {
    for (i = 0; i < 64; i++) {
      /* generate a pseudo-random block of data and upscale accordingly */
      spatial_data[i] = (short) (rand_gen(-L,H) * sign);
    }

    for (k=0; k<NPROPOSAL; k++) {

      if (scale)  for (i=0; i<64; i++) alg_data[k][i] = (spatial_data[i] << scale);
      else        for (i=0; i<64; i++) alg_data[k][i] = spatial_data[i];

      fdct_proposal[k](alg_data[k]);
      idct_proposal[k](alg_data[k]);

      if (scale)  for (i=0; i<64; i++) out_data[k][i] = (alg_data[k][i]+(1<<(scale-1)))>>scale;
      else        for (i=0; i<64; i++) out_data[k][i] = alg_data[k][i];

      for (i=0; i<64; i++) {
        if (spatial_data[i] != out_data[k][i]) {
          delta[k][i] += 1;  
          mismatches[k] += 1;
        }
      }
    }
  }

  fprintf(outputfile, "\n");
  fprintf(outputfile, "Invertibility test:\n");
  fprintf(outputfile, "Test conditions: scale = %d, -L = %ld, +H = %ld, sign = %ld, #iters = %ld \n",
    scale, L, H, sign, run);            
  fprintf(outputfile, "\nProposals: ");
  for (k = 0; k<NPROPOSAL; k++)
    fprintf(outputfile, "%s, ", proposal_label[k]);

  fprintf(outputfile, "\n\nNumber of mismatches:\n");
  for (i=0; i<64; i++)
  {
    fprintf(outputfile, "%2d: ", i) ;
    for (k = 0; k<NPROPOSAL; k++)
    {
      fprintf(outputfile, "%3d, ", delta[k][i]);
    }
    fprintf(outputfile, "\n");	
  }

  fprintf(outputfile, "\n");
  for (k = 0; k<NPROPOSAL; k++)
  {
    if (!mismatches[k]) 
      fprintf(outputfile, "pass, ");
    else
      fprintf(outputfile, "fail, ");
  }
  fprintf(outputfile, "\n") ;		

  return;
}

/* forward DCT test: */
void dct_test (long minpix, long maxpix, long sign, long niters, FILE* outputfile)
{			
  double  max[NPROPOSAL], total[NPROPOSAL], pk[NPROPOSAL], err;
  double  sumerrs[NPROPOSAL][64], sumsqerrs[NPROPOSAL][64];
  double  maxerr[NPROPOSAL][64];
  double  refout[64]; /* output from reference float-point DCT */
  short   ref[64];    /* output from test DCT */
  short   test[64];
  int     proposal, nproposal = NPROPOSAL;
  long    curiter, i; 

  for (proposal = 0; proposal < nproposal; proposal++)
  {
    for(i = 0; i < 64; i++)
      sumerrs[proposal][i] = sumsqerrs[proposal][i] = maxerr[proposal][i] = 0.0;
  }

  /* reset random number generator: */
  rand_gen_reset ();

  // Loop once per generated random-data block
  for (curiter = 0; curiter < niters; curiter++) 
  { 
    // generate a pseudo-random block of data 
    for (i = 0; i < 64; i++) 
    {
      ref[i] = (short) (rand_gen(-minpix,maxpix) * sign);
      refout[i] = (double) ref[i];
    }
    // perform reference forward dct
    ref_double_fdct(refout);
    for (proposal = 0; proposal < nproposal; proposal++)
    {
      memcpy(test, ref, sizeof(short)*64) ;
      fdct_proposal[proposal](test);
      // accumulate error stats 
      for (i = 0; i < 64; i++) 
      {
        err = (double) test[i] - refout[i];
        sumerrs[proposal][i] += err;
        sumsqerrs[proposal][i] += err * err;
        if (err < 0) err = -err;
        if (maxerr[proposal][i] < err) maxerr[proposal][i] = err;
      }
    }
    if (curiter % 100 == 99) {
      fprintf(stderr, ".");
      fflush(stderr);
    }
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "DCT test\n");
  fprintf(outputfile, "Test conditions: -L = %ld, +H = %ld, sign = %ld, #iters = %ld\n",
    minpix, maxpix, sign, niters);
  fprintf(outputfile, "Proposals: ") ;
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", proposal_label[proposal]);
    max[proposal]=total[proposal]=pk[proposal]=0 ;
  }
  fprintf(outputfile,"\n");
  fprintf(outputfile, "Peak absolute values of errors:\n");
  for (i=0; i<64; i++)
  {
    fprintf(outputfile, "%2d: ", i);
    for (proposal = 0; proposal < nproposal; proposal++)
    {
      fprintf(outputfile, "%.2e, ", maxerr[proposal][i]);
      if (pk[proposal] < maxerr[proposal][i]) pk[proposal] = maxerr[proposal][i];
    }
    fprintf(outputfile,"\n");
  }
  fprintf(outputfile, "Worst peak errors: "); 
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%.2e, ", pk[proposal]);
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "(<1.0): ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", meets((double) pk[proposal], 1.0));
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "Mean square errors:\n");
  for (i = 0; i < 64; i++) 
  {
    fprintf(outputfile, "%2d: ", i);
    for (proposal = 0; proposal < nproposal; proposal++)
    {	
      double err = (double) sumsqerrs[proposal][i] / ((double) niters);
      if (max[proposal] < err)
        max[proposal] = err;
      total[proposal] += (double) sumsqerrs[proposal][i];
      fprintf(outputfile, "%.2e, ", ABS(err));
    }
    fprintf(outputfile, "\n");
  }
  fprintf(outputfile, "Worst pmse: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%.2e, ", max[proposal]);
    max[proposal] = 0;
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "\nOverall pmse: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    total[proposal] /= (double) (64*niters);
    fprintf(outputfile, "%.2e, ", total[proposal]);
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "(<0.02): ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", meets(total[proposal], 0.02));
    total[proposal] = 0;
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "\nOverall Mean errors:\n");
  for (i = 0; i < 64; i++) 
  {
    fprintf(outputfile, "%2d: ", i);
    for (proposal = 0; proposal < nproposal; proposal++)
    {
      err = (double) sumerrs[proposal][i] / ((double) niters);
      err = ABS(err);
      total[proposal] += (double) sumerrs[proposal][i];
      fprintf(outputfile, "%.2e, ", err);
      if (max[proposal] < err) max[proposal] = err ;
    }
    fprintf(outputfile, "\n");
  }
  fprintf(outputfile, "\nWorst Mean errors: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%.2e, ", max[proposal]);
    max[proposal] = 0;
  }
  fprintf(outputfile, "\n\nOverall Mean errors: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    total[proposal] /= (double) (64*niters);
    fprintf(outputfile, "%.2e, ", total[proposal]);
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "(<0.0015): ") ;
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", meets(total[proposal], 0.0015));
    total[proposal] = 0 ;
  }
  fprintf(outputfile, "\n") ;
}

void pr_test(long minpix, long maxpix, long sign, long niters, FILE* outputfile)
{
  double max[NPROPOSAL], total[NPROPOSAL], pk[NPROPOSAL], err;
  short  block[64];	    /* random source data */
  short  testout[64];    /* output from test IDCT */
  int    sumerrs[NPROPOSAL][64], sumsqerrs[NPROPOSAL][64];
  int    maxerr[NPROPOSAL][64];
  int    proposal, nproposal = NPROPOSAL ;
  int    i, curiter;

  memset(sumerrs, 0, sizeof(int)*nproposal*64) ;
  memset(sumsqerrs, 0, sizeof(int)*nproposal*64) ;
  memset(maxerr, 0, sizeof(int)*nproposal*64) ;

  /* reset random number generator: */
  rand_gen_reset ();

  // Loop once per generated random-data block 
  for (curiter = 0; curiter < niters; curiter++) 
  {
    // generate a pseudo-random block of data 
    for (i = 0; i < 64; i++)
      block[i] = (short) (rand_gen(-minpix,maxpix) * sign);
    // perform test FDCT 
    for (proposal=0; proposal < nproposal; proposal++)
    {
      memcpy(testout, block, sizeof(short)*64);
      fdct_proposal[proposal](testout) ;
      idct_proposal[proposal](testout) ;
      // perform test IDCT 
      // accumulate error stats 
      for (i = 0; i < 64; i++) {
        err = testout[i] - block[i];
        sumerrs[proposal][i] += (int) err;
        sumsqerrs[proposal][i] += (int) (err * err);
        if (err < 0) err = -err;
        if (maxerr[proposal][i] < err) maxerr[proposal][i] = (int) err;
      }
    }
    if (curiter % 100 == 99) {
      fprintf(stderr, ".");
      fflush(stderr);
    }
  }
  fprintf(outputfile,"\n");
  fprintf(outputfile,"PR Test\n");
  fprintf(outputfile, "Test conditions: -L = %ld, +H = %ld, sign = %ld, #iters = %ld\n",
    minpix, maxpix, sign, niters);
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", proposal_label[proposal]);
    max[proposal] = total[proposal] = pk[proposal]=0;
  }
  fprintf(outputfile, "\n");

  fprintf(outputfile, "Peak absolute values of errors:\n");
  for (i=0; i<64; i++)
  {
    fprintf(outputfile, "%2d: ",i);
    for (proposal = 0; proposal < nproposal; proposal++)
    {
      fprintf(outputfile, "%2d, ", maxerr[proposal][i]);
      if (pk[proposal] < maxerr[proposal][i]) pk[proposal] = maxerr[proposal][i];
    }
    fprintf(outputfile, "\n");
  }
  fprintf(outputfile, "Worst peak error: ") ; 
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%.2e, ", pk[proposal]);
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "(<=1.0): ") ;
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", meets((double) pk[proposal], 1.0));
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "Mean square errors:\n");
  for (i = 0; i < 64; i++) 
  {
    fprintf(outputfile, "%2d: ", i);
    for (proposal = 0; proposal < nproposal; proposal++)
    {	
      double err = (double) sumsqerrs[proposal][i] / ((double) niters);
      if (max[proposal] < err)
        max[proposal] = err;
      total[proposal] += (double) sumsqerrs[proposal][i];
      fprintf(outputfile, "%.2e, ", ABS(err));
    }
    fprintf(outputfile, "\n");
  }
  fprintf(outputfile, "Worst pmse: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%.2e, ", max[proposal]);
    max[proposal]=0 ;
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "\nOverall pmse: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    total[proposal]/= (double) (64*niters);
    fprintf(outputfile, "%.2e, ", total[proposal]);
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "(<0.02): ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", meets(total[proposal], 0.02));
    total[proposal] = 0;
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "\nOverall Mean errors:\n") ;
  for (i = 0; i < 64; i++) 
  {
    fprintf(outputfile, "%2d: ", i) ;
    for (proposal = 0; proposal < nproposal; proposal++)
    {
      err = (double) sumerrs[proposal][i] / ((double) niters);
      err = ABS(err);
      total[proposal] += (double) sumerrs[proposal][i];
      fprintf(outputfile, "%.2e, ", err);
      if (max[proposal] < err) max[proposal] = err;
    }
    fprintf(outputfile, "\n");
  }
  fprintf(outputfile, "\nWorst Mean errors: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%.2e, ", max[proposal]);
    max[proposal] = 0 ;
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "\nOverall Mean errors: ");
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    total[proposal] /= (double) (64*niters);
    fprintf(outputfile, "%.2e, ", total[proposal]);
  }
  fprintf(outputfile, "\n");
  fprintf(outputfile, "(<0.0015): ") ;
  for (proposal = 0; proposal < nproposal; proposal++)
  {
    fprintf(outputfile, "%s, ", meets(total[proposal], 0.0015));
    total[proposal] = 0 ;
  }
  fprintf(outputfile, "\n") ;
  return;
}

void usage ()
{
  printf("ISO/IEC 23002-1 IDCT/DCT precision testbed\n");
  printf("Copyright (C) ISO/IEC 2006\n\n");

  printf("Options: default value inside <...>\n");
  printf("  -l <-256> = L value for pseudo-random test\n");
  printf("  -h <255>  = H value for pseudo-random test\n");
  printf("  -s <1> sign = +1 for normal, -1 to run negated test\n");
  printf("  -i <10000> # iterations \n");
  printf("  -b <8> pixel depth\n");
  printf("  -t <0> test mode:\n"); 
  printf("         0-IDCT, 1-all-zero input, 2-near-DC inversion, 3-DCT, 4-PR test,\n");
  printf("         5-all IDCT tests, 6-all tests, 7-linearity test, 8-invertibility test\n");
  printf("  -m <0> run multiple pseudo-random tests:\n"); 
  printf("         0-none, 1-H.263 tests(+-5,256,300), 2-MPEG tests(+-5...384),\n"); 
  printf("         3-extended dynamic range tests (+-512,1085)\n");
  printf("  -f outputfilename");
  printf("\n");
}

/**************** 
* Runs multiple pseudo-random tests:
*  Mode: 
*    0 - run one explicitly specified test
*    1 - default (H.263) set of tests [Section 5.5.1]
*    2 - above + additional MPEG tests [Annex B, Section B.2]
*    3 - above + additional extended dynamic range tests 
*************
* The extra [-1805,1805] tests are motivated by: 
*   Zhou M. and De Lameillieure J., "IDCT output range in MPEG video coding", 
*   Signal Processing: IMAGE COMMUNICATION, Vol. 11, No. 2,  pp. 137-145,  Dec. 1997.
*/
void run_tests (int mode, 
                long minpix, long maxpix, long sign, long niters, FILE* outputfile,
                void (*test) (long, long, long, long, FILE*))
{
  switch (mode)
  {
  case 0:
    /* just run one particular test: */
    test (minpix, maxpix, sign, niters, outputfile);
    break;
  case 1:
    /* tests specified in section 5.5.1: */
    test (-(1<<B), (1<<B)-1, 1, 10000, outputfile);                 /* -256,255 */
    test (-(1<<B), (1<<B)-1, -1, 10000, outputfile);
    test (-5*(1<<(B-8)), 5*(1<<(B-8)), 1, 10000, outputfile);       /* -5,5 */
    test (-5*(1<<(B-8)), 5*(1<<(B-8)), -1, 10000, outputfile);
    test (-300*(1<<(B-8)), 300*(1<<(B-8)), 1, 10000, outputfile);   /* -300,300 */
    test (-300*(1<<(B-8)), 300*(1<<(B-8)), -1, 10000, outputfile);
    break;
  case 2:
    /* tests specified in section 5.5.1: */
    test (-(1<<B), (1<<B)-1, 1, 10000, outputfile);                 /* -256,255 */
    test (-(1<<B), (1<<B)-1, -1, 10000, outputfile);
    test (-5*(1<<(B-8)), 5*(1<<(B-8)), 1, 10000, outputfile);       /* -5,5 */
    test (-5*(1<<(B-8)), 5*(1<<(B-8)), -1, 10000, outputfile);
    test (-300*(1<<(B-8)), 300*(1<<(B-8)), 1, 10000, outputfile);   /* -300,300 */
    test (-300*(1<<(B-8)), 300*(1<<(B-8)), -1, 10000, outputfile);
    /* additional precision tests specified in Section B.2 */
    test (-(1<<B), (1<<B)-1, 1, 1000000, outputfile);               /* -256,255 */
    test (-(1<<B), (1<<B)-1, -1, 1000000, outputfile);
    test (-5*(1<<(B-8)), 5*(1<<(B-8)), 1, 1000000, outputfile);     /* -5,5 */
    test (-5*(1<<(B-8)), 5*(1<<(B-8)), -1, 1000000, outputfile);
    test (-300*(1<<(B-8)), 300*(1<<(B-8)), 1, 1000000, outputfile); /* -300,300 */
    test (-300*(1<<(B-8)), 300*(1<<(B-8)), -1, 1000000, outputfile);
    test (-384*(1<<(B-8)), 384*(1<<(B-8))-1, 1, 1000000, outputfile); /* -384,383 */
    test (-384*(1<<(B-8)), 384*(1<<(B-8))-1, -1, 1000000, outputfile);
    break;
  case 3:
    /* tests specified in section 5.5.1: */
    test (-(1<<B), (1<<B)-1, 1, 10000, outputfile);                 /* -256,255 */
    test (-(1<<B), (1<<B)-1, -1, 10000, outputfile);
    test (-5*(1<<(B-8)), 5*(1<<(B-8)), 1, 10000, outputfile);       /* -5,5 */
    test (-5*(1<<(B-8)), 5*(1<<(B-8)), -1, 10000, outputfile);
    test (-300*(1<<(B-8)), 300*(1<<(B-8)), 1, 10000, outputfile);   /* -300,300 */
    test (-300*(1<<(B-8)), 300*(1<<(B-8)), -1, 10000, outputfile);
    /* extra dynamic range tests with Q=10000: */
    test (-384*(1<<(B-8)), 384*(1<<(B-8))-1, 1, 10000, outputfile); /* -384,383 */
    test (-384*(1<<(B-8)), 384*(1<<(B-8))-1, -1, 10000, outputfile);
    test (-512*(1<<(B-8)), 512*(1<<(B-8))-1, 1, 10000, outputfile); /* -512,511 */
    test (-512*(1<<(B-8)), 512*(1<<(B-8))-1, -1, 10000, outputfile);
    test (-512*(1<<(B-8)), 512*(1<<(B-8))-1, 1, 10000, outputfile); /* -512,511 */
    test (-512*(1<<(B-8)), 512*(1<<(B-8))-1, -1, 10000, outputfile);
    test (-1805*(1<<(B-8)), 1805*(1<<(B-8)), 1, 10000, outputfile); /* -1805,1805 */
    test (-1805*(1<<(B-8)), 1805*(1<<(B-8)), -1, 10000, outputfile);
    /* additional precision tests specified in Section B.2 */
    test (-(1<<B), (1<<B)-1, 1, 1000000, outputfile);               /* -256,255 */
    test (-(1<<B), (1<<B)-1, -1, 1000000, outputfile);
    test (-5*(1<<(B-8)), 5*(1<<(B-8)), 1, 1000000, outputfile);     /* -5,5 */
    test (-5*(1<<(B-8)), 5*(1<<(B-8)), -1, 1000000, outputfile);
    test (-300*(1<<(B-8)), 300*(1<<(B-8)), 1, 1000000, outputfile); /* -300,300 */
    test (-300*(1<<(B-8)), 300*(1<<(B-8)), -1, 1000000, outputfile);
    test (-384*(1<<(B-8)), 384*(1<<(B-8))-1, 1, 1000000, outputfile); /* -384,383 */
    test (-384*(1<<(B-8)), 384*(1<<(B-8))-1, -1, 1000000, outputfile);
    /* extra dynamic range tests with Q=1000000: */
    test (-512*(1<<(B-8)), 512*(1<<(B-8))-1, 1, 1000000, outputfile); /* -512,511 */
    test (-512*(1<<(B-8)), 512*(1<<(B-8))-1, -1, 1000000, outputfile);
    test (-1805*(1<<(B-8)), 1805*(1<<(B-8)), 1, 1000000, outputfile); /* -1805,1805 */
    test (-1805*(1<<(B-8)), 1805*(1<<(B-8)), -1, 1000000, outputfile);
  }
}

/* main program: */
int main (int argc, char *argv[])
{
  int i = 1;
  int minpix = -256;
  int maxpix = 255; 
  int sign = 1; 
  int niters = 10000;
  int test_mode = 5;
  int m = 0;
  FILE *outputfile = NULL;
  char *outputfilename = NULL;

  if (argc <= 1)
  {
    usage ();
    return 0;
  }

  /* parse arguments: */
  while (i < argc && argv[i][0] == '-')
  {
    switch (argv[i][1])
    {
    case 'l': case 'L':
      minpix = atoi (argv[i+1]);
      break;
    case 'h': case 'H':
      maxpix = atoi (argv[i+1]);
      break;
    case 'i':
      niters = atoi (argv[i+1]);
      break;
    case 's':
      sign = atoi (argv[i+1]);
      break;
    case 'b': case 'B':
      B = atoi (argv[i+1]);
      break;
    case 't':
      test_mode = atoi (argv[i+1]);
      break;
    case 'm':
      m = atoi (argv[i+1]);
      break;
    case 'f':
      outputfilename = argv[i+1];
      break;
    case 'e':
      scale = atoi(argv[i+1]);
      break;
    case 'u': case 'U': case '?': 
    default:
      usage ();
      return 1;
    }
    i += 2;  
  }

  if (outputfilename)
  {
    if((outputfile = fopen(outputfilename, "wt")) == NULL)
    {
      fprintf(outputfile, "\nError openining output file!");
      exit(-1);
    }
  }
  else
    outputfile = stderr;

  dct_init();

  switch(test_mode)
  {
  case 0:
    run_tests (m, minpix, maxpix, sign, niters, outputfile, idct_test);		
    break;
  case 1:
    zeroinput_test(outputfile);
    break;
  case 2:
    neardc_test(outputfile);
    break;
  case 3:
    run_tests (m, minpix, maxpix, sign, niters, outputfile, dct_test);		
    break;
  case 4:
    run_tests (m, minpix, maxpix, sign, niters, outputfile, pr_test);
    break;

  case 5: // run all idct test
    run_tests (m, minpix, maxpix, sign, niters, outputfile, idct_test);
    zeroinput_test (outputfile);
    neardc_test (outputfile);
    linearity_test (outputfile); //Lijie Liu
    break ;

  case 6: // run all test
    run_tests (m, minpix, maxpix, sign, niters, outputfile, idct_test);
    zeroinput_test (outputfile);
    neardc_test (outputfile);
    run_tests (m, minpix, maxpix, sign, niters, outputfile, dct_test);	
    run_tests (m, minpix, maxpix, sign, niters, outputfile, pr_test);
    linearity_test (outputfile); //cixunzhang
    run_tests (m, minpix, maxpix, sign, niters, outputfile, invert_test); // arianne
    break ;

  case 7: //run linearity test
    linearity_test (outputfile);
    break;

  case 8: 
    run_tests (m, minpix, maxpix, sign, niters, outputfile, invert_test); // arianne
    break;

  default:
    fprintf(outputfile, "mode not supported\n");
    break;
  }
  if (outputfilename)
    fclose(outputfile);

  return 0;
}

